from pandas import DataFrame, concat
import os
from itertools import product
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '4'
import shutil
import glob
from tensorflow.keras.layers import Dense, LSTM, Input, BatchNormalization, Conv1D, MaxPooling1D, Dropout, Flatten
from keras.regularizers import l1, l2, l1_l2
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import Model
from tensorflow.keras.models import model_from_json
from kerastuner.tuners import RandomSearch
from kerastuner.engine.hyperparameters import HyperParameters

try:
    from functions import *
    from configs import hyper_conf, accept_threshold_for_loss_diff, parameter_tuning_trials
    from data_access import *
except Exception as e:
    from .functions import *
    from .configs import hyper_conf, accept_threshold_for_loss_diff, parameter_tuning_trials
    from .data_access import *
    from .utils import get_current_day


def model_from_to_json(path=None, weights_path=None, model=None, is_writing=False):
    if is_writing:
        model_json = model.to_json()
        with open(path, "w") as json_file:
            json_file.write(model_json)
        model.save_weights(weights_path)
    else:
        json_file = open(path, 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        model = model_from_json(loaded_model_json)
        model.load_weights(weights_path)
        return model


def get_params(params, comb):
    count = 0
    for p in params:
        _p = type(params[p])(comb[count])
        params[p] = _p
        count += 1
    return params


def get_order_freq(predicted_orders, customer_indicator, time_indicator):
    if len(predicted_orders) != 0:
        predicted_orders['order_seq_num'] = predicted_orders.sort_values(
            by=[customer_indicator, time_indicator]).groupby([customer_indicator]).cumcount() + 1
    return predicted_orders


def get_number_of_orders(predicted_orders, customer_indicator):
    number_of_orders = None
    if len(predicted_orders) != 0:
        number_of_orders = predicted_orders.groupby(customer_indicator).agg({"order_seq_num": "max"}).reset_index()
    return number_of_orders


class TrainConv1Dimension:
    def __init__(self,
                 date=None,
                 time_indicator=None,
                 order_count=None,
                 data_source=None,
                 data_query_path=None,
                 time_period=None,
                 directory=None,
                 customer_indicator=None,
                 predicted_orders=None,
                 amount_indicator=None):
        self.directory = directory
        self.time_indicator = time_indicator
        self.customer_indicator = customer_indicator
        self.order_count = order_count
        self.time_period = time_period
        self.amount_indicator = amount_indicator
        self.params = hyper_conf('purchase_amount')
        self.hyper_params = get_tuning_params(hyper_conf('purchase_amount_hyper'), self.params)
        self.optimized_parameters = {}
        self._p = None
        self.data, self.features, self.y, self.c_min_max, \
        self.max_date = data_manipulation(
                                           date=date,
                                           amount_indicator=amount_indicator,
                                           time_indicator=time_indicator,
                                           order_count=order_count,
                                           data_source=data_source,
                                           data_query_path=data_query_path,
                                           customer_indicator=customer_indicator,
                                           directory=self.directory)
        self.params['feature_count'] = self.y
        self.hp = HyperParameters()
        self.model = check_model_exists(self.directory, "trained_purchase_amount_model", self.time_period)
        self.input = None
        self.model_data = {}
        self.result = None
        self.residuals, self.anomaly = [], []
        self.customers = list(self.data[customer_indicator].unique())
        self.predicted_orders = get_order_freq(predicted_orders, customer_indicator, time_indicator)
        self.num_of_future_orders = get_number_of_orders(self.predicted_orders, self.customer_indicator)
        self.results = DataFrame()

    def train_test_split(self):
        self.model_data['train'], self.model_data['test'] = random_data_split(self.data, float(self.params['split_ratio']))

    def data_preparation(self, is_prediction=False):
        if not is_prediction:
            self.train_test_split()
        else:
            self.model_data['prediction'] = self.data
        self.model_data = reshape_data(self.model_data, self.features, self.y, is_prediction)

    def build_parameter_tuning_model(self, hp):
        self.input = Input(shape=(self.model_data['x_train'].shape[1], 1,))
        ### conv 1D layer
        conv = Conv1D(filters=hp.Choice('filters', self.hyper_params['filters']),
                      kernel_size=hp.Choice('kernel_size', self.hyper_params['kernel_size']),
                      padding='same',
                      activation=hp.Choice('activation', self.hyper_params['activation']),
                      kernel_regularizer=l1_l2(l1=hp.Choice('l1', self.hyper_params['l1']),
                                               l2=hp.Choice('l2', self.hyper_params['l2'])),
                      bias_regularizer=l2(hp.Choice('l2', self.hyper_params['l2'])),
                      activity_regularizer=l2(hp.Choice('l2', self.hyper_params['l2']))
                      )(self.input)
        conv = BatchNormalization()(conv)
        # LSTM layer
        conv = MaxPooling1D(hp.Choice('max_pooling_unit', self.hyper_params['max_pooling_unit']))(conv)
        conv = Dropout(hp.Choice('drop_out_ratio', self.hyper_params['drop_out_ratio']))(conv)
        conv = LSTM(hp.Choice('lstm_units', self.hyper_params['lstm_units']),
                    use_bias=False,
                    activation=hp.Choice('activation', self.hyper_params['activation']),
                    kernel_regularizer=l1_l2(l1=hp.Choice('l1', self.hyper_params['l1']),
                                             l2=hp.Choice('l2', self.hyper_params['l2'])),
                    bias_regularizer=l2(hp.Choice('l2', self.hyper_params['l2'])),
                    activity_regularizer=l2(hp.Choice('l2', self.hyper_params['l2']))
                    )(conv)
        conv = BatchNormalization()(conv)
        conv = Flatten()(conv)
        # layers after flattened layers
        for i in range(hp.Int('num_layers',
                              self.hyper_params['num_layers']['min'],
                              self.hyper_params['num_layers']['max'])):
            conv = Dense(hp.Choice('units', self.hyper_params['units']),
                         hp.Choice('activation', self.hyper_params['activation']),
                         kernel_regularizer=l1_l2(l1=hp.Choice('l1', self.hyper_params['l1']),
                                                  l2=hp.Choice('l2', self.hyper_params['l2'])),
                         bias_regularizer=l2(hp.Choice('l2', self.hyper_params['l2'])),
                         activity_regularizer=l2(hp.Choice('l2', self.hyper_params['l2']))
                         )(conv)
            conv = BatchNormalization()(conv)
        output = Dense(1,
                       activation=hp.Choice('activation', self.hyper_params['activation']),
                       kernel_regularizer=l1_l2(l1=hp.Choice('l1', self.hyper_params['l1']),
                                                l2=hp.Choice('l2', self.hyper_params['l2'])),
                       bias_regularizer=l2(hp.Choice('l2', self.hyper_params['l2'])),
                       activity_regularizer=l2(hp.Choice('l2', self.hyper_params['l2']))
                       )(conv)
        model = Model(inputs=self.input, outputs=output)
        model.compile(loss=self.params['loss'],
                      optimizer=Adam(lr=hp.Choice('lr', self.hyper_params['lr'])),
                      metrics=[hp.Choice('loss', self.hyper_params['loss'])])
        return model

    def build_model(self):
        print(self.params)
        self.input = Input(shape=(self.model_data['x_train'].shape[1], 1,))
        # conv 1D layer
        conv = Conv1D(filters=self.params['filters'],
                      kernel_size=self.params['kernel_size'],
                      padding='same',
                      activation=self.params['activation'],
                      kernel_regularizer=l1_l2(l1=self.params['l1'],
                                               l2=self.params['l2']),
                      bias_regularizer=l2(self.params['l2']),
                      activity_regularizer=l2(self.params['l2'])
                      )(self.input)
        conv = BatchNormalization()(conv)
        # LSTM layer
        conv = MaxPooling1D(self.params['max_pooling_unit'])(conv)
        conv = Dropout(self.params['drop_out_ratio'])(conv)
        conv = LSTM(self.params['lstm_units'],
                    use_bias=False,
                    activation=self.params['activation'],
                    kernel_regularizer=l1_l2(l1=self.params['l1'],
                                             l2=self.params['l2']),
                    bias_regularizer=l2(self.params['l2']),
                    activity_regularizer=l2(self.params['l2'])
                    )(conv)
        conv = BatchNormalization()(conv)
        conv = Flatten()(conv)
        # layers after flattened layers
        for i in range(self.params['num_layers']):
            conv = Dense(self.params['units'],
                         self.params['activation'],
                         kernel_regularizer=l1_l2(l1=self.params['l1'],
                                                  l2=self.params['l2']),
                         bias_regularizer=l2(self.params['l2']),
                         activity_regularizer=l2(self.params['l2'])
                         )(conv)
            conv = BatchNormalization()(conv)
        output = Dense(1,
                       activation=self.params['activation'],
                       kernel_regularizer=l1_l2(l1=self.params['l1'],
                                                l2=self.params['l2']),
                       bias_regularizer=l2(self.params['l2']),
                       activity_regularizer=l2(self.params['l2'])
                       )(conv)
        self.model = Model(inputs=self.input, outputs=output)
        self.model.compile(loss=self.params['loss'],
                           optimizer=Adam(lr=self.params['lr']),
                           metrics=[self.params['loss']])

    def learning_process(self, save_model=True, history=True):
        if history:
            history = self.model.fit(self.model_data['x_train'],
                                     self.model_data['y_train'],
                                     batch_size=int(self.params['batch_size']),
                                     epochs=int(self.params['epochs']),
                                     verbose=1,
                                     validation_data=(self.model_data['x_test'], self.model_data['y_test']),
                                     shuffle=True)
        else:
            print("*" * 5, "Fit Purchase Amount Model", "*" * 5)
            self.model.fit(self.model_data['x_train'],
                           self.model_data['y_train'],
                           batch_size=int(self.params['batch_size']),
                           epochs=int(self.params['epochs']),
                           verbose=1,
                           validation_data=(self.model_data['x_test'], self.model_data['y_test']),
                           shuffle=True)

        if save_model:
            model_from_to_json(path=model_path(self.directory,
                                               "trained_purchase_amount_model", self.time_period),
                               weights_path = weights_path(self.directory,
                                               "trained_purchase_amount_model", self.time_period),
                               model=self.model,
                               is_writing=True)

        if history:
            return history

    def train_execute(self):
        print("*" * 5, " Purchase Amount train model process ", "*" * 5)
        print(self.model)
        if self.model is None:
            self.data_preparation()
            self.parameter_tuning()
            self.build_model()
            self.learning_process()
        else:
            self.model = model_from_to_json(path=model_path(self.directory,
                                                            "trained_purchase_amount_model", self.time_period),
                                            weights_path=weights_path(self.directory,
                                                                      "trained_purchase_amount_model", self.time_period))
            print("Previous model already exits in the given directory  '" + self.directory + "'.")

    def prediction_per_customer(self, customer):
        _prediction_data = pd.DataFrame(prediction_data[customer]['data'])
        _number = prediction_data[customer]['number']
        _user_min = prediction_data[customer]['user_min']
        _user_max = prediction_data[customer]['user_max']
        _prediction = get_prediction(_prediction_data,
                                     _number,
                                     self.model.input.shape[1],
                                     self.model)

        prediction_data[customer]['prediction'] = get_predicted_data_readable_form(customer,
                                                                                   _prediction,
                                                                                   self.model.input.shape[1] + 1,
                                                                                   _user_min,
                                                                                   _user_max,
                                                                                   self.customer_indicator)

    def rearrange_prediction_data(self):
        self.model_data['prediction_data'] = pd.merge(self.model_data['prediction_data'],
                                                      self.num_of_future_orders[[self.customer_indicator, 'order_seq_num']],
                                                      on=self.customer_indicator, how='left')
        self.model_data['prediction_data'] = self.model_data['prediction_data'].query("order_seq_num == order_seq_num")
        self.model_data['prediction_data'] = self.model_data['prediction_data'].sort_values(by=self.customer_indicator)
        self.customers = list(self.model_data['prediction_data'][self.model_data['prediction_data'].isin(
            list(self.num_of_future_orders[self.customer_indicator].unique()))][self.customer_indicator].unique())

    def get_user_of_historic_data(self, sample_customers):
        _sample_p_data = self.model_data['prediction_data'][
            self.model_data['prediction_data'][self.customer_indicator].isin(sample_customers)]
        _c_min_max = np.array(self.c_min_max.sort_values(by=self.customer_indicator)[['user_min', 'user_max']]).tolist()
        _numbers = list(_sample_p_data['order_seq_num'])
        _historic_data = _sample_p_data.drop([self.customer_indicator, 'order_seq_num'], axis=1).to_dict('results')
        _values = list(zip(sample_customers, _historic_data, _numbers, _c_min_max))
        return {u[0]: {"data": [u[1]],
                       "number": u[2],
                       'user_min': u[3][0],
                       'user_max': u[3][1]} for u in _values}

    def parallel_prediction(self):
        try:
            os.mkdir(join(self.directory, "temp_purchase_amount_results", ""))
        except Exception as e:
            print(e)
            print("recreating 'temp_purchase_amount_results' folder ...")
            shutil.rmtree(join(self.directory, "temp_purchase_amount_results", ""))
            os.mkdir(join(self.directory, "temp_purchase_amount_results", ""))

        iters = int(len(self.customers) / 256) + 1
        print("number of iters :", iters)
        for i in range(iters):
            print("main iteration :", str(i), " / ", str(iters))
            _sample_customers = get_iter_sample(self.customers, i, iters, 256)
            global prediction_data
            prediction_data = self.get_user_of_historic_data(_sample_customers)
            execute_parallel_run(_sample_customers, self.prediction_per_customer, arguments=None, parallel=8)
            _result = pd.DataFrame()
            for c in _sample_customers:
                try:
                    _result = concat([_result, prediction_data[c]['prediction']])
                except Exception as e:
                    print(c)
            _result.to_csv(join(self.directory, "temp_purchase_amount_results", str(_sample_customers[0] + "_data.csv")),
                           index=False)

    def create_prediction_result_data(self):
        print("merge predicted data ...")
        _import_files = glob.glob(join(self.directory, "temp_purchase_amount_results", "*.csv"))
        li = []
        for f in _import_files:
            try:
                _df = pd.read_csv(f, index_col=None, header=0)
                li.append(_df)
            except Exception as e:
                print(e)
        self.results = concat(li, axis=0, ignore_index=True)
        # shutil.rmtree(join(self.directory, "temp_purchase_amount_results", ""))
        print(self.results.head())

    def prediction_execute(self):
        print("number of users :", len(self.customers))
        self.model_data['prediction_data'] = self.data[self.features + [self.customer_indicator]]
        if self.model is not None:
            self.model = model_from_to_json(path=model_path(self.directory,
                                                            "trained_purchase_amount_model", self.time_period),
                                            weights_path=weights_path(self.directory,
                                                                      "trained_purchase_amount_model", self.time_period))

        if self.num_of_future_orders is not None:
            print(len(self.customers))
            self.rearrange_prediction_data()
            print(len(self.customers))
            self.parallel_prediction()
            self.create_prediction_result_data()
        else:
            self.results = merge_0_result_at_time_period(self.data,
                                                         self.max_date,
                                                         self.time_period,
                                                         self.customer_indicator,
                                                         self.time_indicator,
                                                         self.amount_indicator)
        self.results = merging_predicted_date_to_result_data(self.results,
                                                             self.predicted_orders,
                                                             self.customer_indicator,
                                                             self.time_indicator,
                                                             self.amount_indicator)
        self.results = check_for_previous_predicted_clv_results(self.results,
                                                                self.directory,
                                                                self.time_period,
                                                                self.time_indicator,
                                                                self.customer_indicator,
                                                                )
        self.results.to_csv(get_result_data_path(self.directory, self.time_period, self.max_date), index=False)
        print("lifetime value :")
        print(self.results.head())

    def initialize_keras_tuner(self):
        """
        Parameter tuning process is triggered via Keras-Turner Library.
        However, batch_size and epoch parameters of optimization are created individually.
        :return:
        """

        tuner = RandomSearch(
            self.build_parameter_tuning_model,
            max_trials=parameter_tuning_trials,
            hyperparameters=self.hp,
            allow_new_entries=True,
            objective='loss')
        tuner.search(x=self.model_data['x_train'],
                     y=self.model_data['y_train'],
                     epochs=5,
                     verbose=1,
                     validation_data=(self.model_data['x_test'], self.model_data['y_test']))

        for p in tuner.get_best_hyperparameters()[0].values:
            if p in list(self.params.keys()):
                self.params[p] = tuner.get_best_hyperparameters()[0].values[p]

    def batch_size_and_epoch_tuning(self):
        """
        It finds the optimum batch_size and epoch.
        Each chosen epoch and batch_size of model created via model.
        Last epoch of loss is lower than 'accept_threshold_for_loss_diff' decision be finalized.
        Epoch will test ascending format (e.g. 4, 8, 16, ..., 1024) in order to minimize time consumption
        of both parameter tuning and model creation.
        Batch size will test descending format (e.g. 1024, 512, ...4) in order to minimize time consumption
        of both parameter tuning and model creation.
        """

        counter = 0
        optimum_epoch_process_done = False
        epoch_bs_combs = list(product(sorted(self.hyper_params['batch_sizes'], reverse=True),
                                      sorted(self.hyper_params['epochs'])
                                      ))
        loss_values = []
        while not optimum_epoch_process_done:
            self.params['epochs'] = int(epoch_bs_combs[counter][1])
            self.params['batch_size'] = int(epoch_bs_combs[counter][0])
            self.build_model()
            _history = self.learning_process(save_model=False, history=True)
            if _history.history['loss'][-1] < accept_threshold_for_loss_diff:
                optimum_epoch_process_done = True
            loss_values.append({"epochs": self.params['epochs'],
                                "batch_size": self.params['batch_size'],
                                "loss": _history.history['loss'][-1]})
            counter += 1
            if counter >= parameter_tuning_trials:
                optimum_epoch_process_done = True
                loss_values_df = pd.DataFrame(loss_values).sort_values(by='loss', ascending=True)
                self.params['epochs'] = list(loss_values_df['epochs'])[0]
                self.params['batch_size'] = list(loss_values_df['batch_size'])[0]

    def remove_keras_tuner_folder(self):
        """
        removing keras tuner file. while you need to update the parameters it will affect rerun the parameter tuning.
        It won`t start unless the folder has been removed.
        """

        try:
            shutil.rmtree(
                join(abspath(__file__).split("purchase_amount_model.py")[0].split("clv")[0][:-1], "clv_prediction",
                     "untitled_project"))
        except Exception as e:
            print(" Parameter Tuning Keras Turner dummy files have already removed!!")

    def update_tuned_parameter_file(self):
        """
        tuned parameters are stored at 'test_parameters.yaml.' in given 'export_path' argument.
        """

        try:
            _params = read_yaml(self.directory, "test_parameters.yaml")
        except Exception as e:
            print(e)
            _params = None
        _params['purchase_amount'] = self.params if _params is not None else {'purchase_amount': self.params}
        write_yaml(self.directory, "test_parameters.yaml", _params, ignoring_aliases=True)

    def parameter_tuning(self):
        if check_for_existing_parameters(self.directory, 'purchase_amount') is None:
            self.initialize_keras_tuner()
            self.batch_size_and_epoch_tuning()
            self.remove_keras_tuner_folder()
            self.update_tuned_parameter_file()
        else:
            self.params = check_for_existing_parameters(self.directory, 'purchase_amount')




import numpy as np
import pandas as pd
from math import sqrt
import random
from os import listdir
from os.path import dirname
from dateutil.parser import parse
from statsmodels.tsa.arima.model import ARIMA

try:
    from data_access import GetData
    from utils import *
    from configs import accepted_ratio_of_actual_order
except Exception as e:
    from .data_access import GetData
    from .utils import *
    from .configs import accepted_ratio_of_actual_order


def data_manipulation_np(date,
                         time_indicator,
                         data_source,
                         data_query_path,
                         feature,
                         customer_indicator,
                         params,
                         time_period,
                         directory):
    data_process = GetData(data_source=data_source,
                           data_query_path=data_query_path,
                           time_indicator=time_indicator,
                           feature=feature, date=date)
    data_process.data_execute()
    print("data size :", len(data_process.data))
    data = data_process.data
    data[time_indicator] = data[time_indicator].apply(lambda x: convert_str_to_day(x))
    data['last_days'] = data.sort_values(by=[customer_indicator, time_indicator],
                                         ascending=True).groupby(customer_indicator)[time_indicator].shift(1)
    data = data.query("last_days == last_days")
    data = pd.merge(data, data.rename(columns={"last_days": "last_days_2"}).groupby(
                                            customer_indicator)['last_days_2'].max(),
                    on=customer_indicator, how='left')
    data['time_diff'] = data.apply(lambda row: calculate_time_diff(row['last_days'],
                                                                   row[time_indicator],
                                                                   time_period), axis=1)
    opt_lag = OptimumLagDecision(data, customer_indicator, time_indicator, params, directory)
    opt_lag.find_optimum_lag()
    params['lahead'], params['lag'] = opt_lag.best_lag, opt_lag.best_lag
    params['batch_size'] = max((params['tsteps'] - 1), (params['lahead'] - 1)) + (params['lahead'] * 2)

    data, customer_min_max = get_customer_min_max_data(data, 'time_diff', customer_indicator)
    return data, 'time_diff_norm', customer_min_max, params


def data_manipulation(date,
                      time_indicator,
                      order_count,
                      data_source,
                      data_query_path,
                      amount_indicator,
                      customer_indicator,
                      directory):
    data_process = GetData(data_source=data_source,
                           data_query_path=data_query_path,
                           time_indicator=time_indicator,
                           feature=amount_indicator,
                           date=date)
    data_process.data_execute()
    print("data size :", len(data_process.data))
    data = data_process.data
    data[time_indicator] = data[time_indicator].apply(lambda x: convert_str_to_day(x))
    max_date = max(data[time_indicator])
    data = data.sort_values(by=[customer_indicator, time_indicator], ascending=True)
    data['order_seq_num'] = data.sort_values(by=[customer_indicator,
                                                 time_indicator]).groupby([customer_indicator]).cumcount() + 1
    data = data.sort_values(by=[customer_indicator, time_indicator])
    data = pd.merge(data,
                    data.groupby(customer_indicator)['order_seq_num'].max().reset_index().rename(
                        columns={"order_seq_num": "max_order"}),
                    on=customer_indicator, how='left')
    order_count = order_count_decision(data, order_count, customer_indicator, directory)
    data['prev_orders'] = data['max_order'] - order_count
    data = data.query("order_seq_num > prev_orders")
    data['order_seq_num'] = data.sort_values(by=[customer_indicator, time_indicator]).groupby(
        [customer_indicator]).cumcount() + 1
    data['order_seq_num'] = data.apply(
        lambda row: row['order_seq_num'] + abs(row['prev_orders']) if row['prev_orders'] < 0 else row['order_seq_num'],
        axis=1)
    data, customer_min_max = get_customer_min_max_data(data, amount_indicator, customer_indicator)
    data = pivoting_orders_sequence(data, customer_indicator, amount_indicator)
    features = list(range(1, order_count))
    return data, features, order_count, customer_min_max, max_date


def order_count_decision(data, order_count, customer_indicator, directory):
    """
    This allows us to calculate an optimum number of order counts.
    Order count is also the feature number of the purchase amount model.

    !!! Caution !!!
        - Order count must be inserted into the test_parameters.yaml in order not to allow for changing later on prediction.
        - Once the model is built with calculated order_count it must be predicted with the same order count.

    !!! Why we need an order count of a decision? !!!
        - it is a crucial parameter for the purchase amount model.
        - The purchase amount is a 1 Dimensional Conv NN. It works with kernels and it is sizes are related to feature size.
        - At the purchase amount model features are sequential orders.
        - For instance if we assign order count as 5, user_1, user_2, user_3, user_4 have 100, 101, 300, 2 orders.
            The data set will be;
                - user_1: 95th, 96th, 97th, 98th, 99th, 100th  orders
                - user_2: 96th, 97th, 98th, 99th, q00th, 101st  orders
                - user_3: 295th, 296th, 297th, 298th, 299th, 300th  orders
                - user_4: only have 2 orders first 9 orders will be 0 and this will affect the model process.
        - That is why it is crucial to have a minimum 0 assigned order as user_4
          However, it is also a crucial point  to get as much previous order count for make kernel size larger,

    !!! How does it work? !!!
        - At least the last 5 orders are accepted for the model. Lower than 5 orders will not be fully Dimensional Conv.NN.
          It will work as Feed Forward NN because of the lack of kernel size.
        - Iterate from 5 to max order count for any customer at the data set.
        - Finds zero assign number of orders
        - Calculate expected total numbers with given order count; order_count * unique_customers
        - Calculate ratio: zero assign number of orders / expected total numbers with given order count
        - Collect the order count which has ratio less than 0.05
        - Assign as order count form max collected order count (ratio < 0.05)


    :param data: raw data with order_sequential per customer
    :param order_count: if order_count is assigned when the platform is triggered.
    :param customer_indicator: customer columns on data frame
    :param directory: need to check order count has been decided before on test_parameters.yaml
    :return: decided order_count
    """
    if order_count is None:
        params = check_for_existing_parameters(directory, 'purchase_amount')
        if params is None:
            total_orders = []
            max_order = max(data['max_order'])
            data = data.drop('max_order', axis=1)
            for rc in range(5, max_order):
                _data = pd.merge(data,
                                 data.groupby(customer_indicator)['order_seq_num'].max().reset_index().rename(
                                     columns={"order_seq_num": "max_order"}),
                                 on=customer_indicator, how='left')
                _data['prev_orders'] = _data['max_order'] - rc
                _data = _data.query("order_seq_num > prev_orders")
                _total_unique_customers = len(_data[customer_indicator].unique())
                _total_orders = len(_data)
                total_orders.append({"order_count": rc,
                                     "total_orders": _total_orders,
                                     "total_data_point": _total_unique_customers * rc,
                                     "zero_orders": (_total_unique_customers * rc) - _total_orders,
                                     "ratio": _total_orders / (_total_unique_customers * rc),
                                     'unique_customers': _total_unique_customers})
            total_orders = pd.DataFrame(total_orders)
            df = total_orders.query("ratio >= @accepted_ratio_of_actual_order")
            df = df.sort_values(by='ratio', ascending=False)
            df = df.sort_values(by='order_count', ascending=False)
            try:
              order_count = list(df['order_count'])[0]
            except Exception as e:
              print(e)
              order_count = list(total_orders['order_count'])[0]
        else:
            order_count = params['feature_count']
    return order_count


def min_max_norm(value, _min, _max):
    if abs(_max - _min) != 0:
        return (value - _min) / abs(_max - _min)
    else: return 0


def get_customer_min_max_data(data, feature, customer_indicator):
    data['user_max'], data['user_min'] = data[feature], data[feature]
    users_min_max = data.groupby(customer_indicator).agg({"user_max": "max", "user_min": "min"}).reset_index()
    data = pd.merge(data.drop(["user_max", "user_min"], axis=1), users_min_max, on=customer_indicator, how='left')
    data[feature + '_norm'] = data.apply(lambda row: min_max_norm(row[feature],
                                                                  row['user_min'],
                                                                  row['user_max']), axis=1)
    return data, users_min_max


def pivoting_orders_sequence(data, customer_indicator, feature):
    data = pd.DataFrame(np.array(data.pivot_table(index=customer_indicator,
                                                  columns="order_seq_num",
                                                  aggfunc={feature + "_norm": "first"}
                                                  ).reset_index())).rename(columns={0: customer_indicator})
    data = data.fillna(0)
    return data


def calculate_time_diff(date, prev_date, time_period):
    date = datetime.datetime.strptime(str(date)[0:19], '%Y-%m-%d %H:%M:%S')
    prev_date = datetime.datetime.strptime(str(prev_date)[0:19], '%Y-%m-%d %H:%M:%S')
    if time_period == 'hour':
        return abs((date - prev_date).total_seconds()) / 60
    if time_period == 'day':
        return abs((date - prev_date).total_seconds()) / 60 / 60
    if time_period not in ['hour', 'day']:
        return abs((date - prev_date).total_seconds()) / 60 / 60 / 24


def sampling(sample, sample_size):
    if len(sample) <= sample_size:
        return sample
    else:
        return random.sample(sample, sample_size)


def get_sample_size(ratio, sample):
    return int(ratio * len(sample))


def random_data_split(data, ratio):
    index = range(len(data))
    train_index = random.sample(index, int(len(data) * ratio))
    test_index = list(set(index) - set(train_index))
    train = data.iloc[train_index]
    test = data.iloc[test_index]
    return train, test


def reshape_data(model_data, features, y, prediction):
    if not prediction:
        model_data['x_train'] = model_data['train'][features].values.reshape(
            len(model_data['train']), len(features), 1)
        model_data['y_train'] = model_data['train'][y].values
        model_data['x_test'] = model_data['test'][features].values.reshape(
            len(model_data['test']),len(features), 1)
        model_data['y_test'] = model_data['test'][y].values
    else:
        model_data['prediction_x'] = model_data['prediction'][features].values.reshape(
            len(model_data['prediction']), len(features), 1)
    return model_data


def data_for_customer_prediction(data, prediction_data, params):
    if len(prediction_data) != 0:
        data = pd.concat([data[['time_diff_norm']], prediction_data[['time_diff_norm']]])
    x = pd.DataFrame(np.repeat(data[['time_diff_norm']].values, repeats=params['lag'], axis=1))
    shift_day = int(params['lahead'] / params['lag'])
    if params['lahead'] > 1:
        for i, c in enumerate(x.columns):
            x[c] = x[c].shift(i * shift_day)  # every each same days of shifted
    to_drop = max((params['tsteps'] - 1), (params['lahead'] - 1))
    return reshape_3(x[to_drop:].values)


def check_for_next_prediction(data, model_num):
    columns = sorted([int(col) for col in data.columns], reverse=False)[-model_num:]
    data_for_pred = pd.DataFrame([list(data[col])[0] for col in columns])
    data_for_pred = data_for_pred.values.reshape(1, model_num, 1)
    return data_for_pred


def add_predicted_values_as_column(data, pred):
    max_num = max(list(data.columns))
    data[max_num + 1] = pred
    return data


def get_prediction(data, number, model_num, model):
    for num in range(0, int(number) + 1):
        _pred_data = check_for_next_prediction(data, model_num)
        _pred = model.predict(_pred_data)[0]
        data = add_predicted_values_as_column(data, _pred)
    return data


def get_predicted_data_readable_form(user, prediction, removing_columns, _user_min, _user_max, customer_indicator):
    removing_cols = list(range(removing_columns + 1))
    predictions = [{customer_indicator: user,
                    "user_min": _user_min,
                    "user_max": _user_max,
                    "pred_order_seq": col - removing_columns,
                    "prediction": list(prediction[col])[0]} for col in prediction.columns if col not in removing_cols]
    predictions = pd.DataFrame(predictions)
    predictions['prediction_values'] = predictions.apply(
        lambda row: ((row['user_max'] - row['user_min']) * row['prediction']) + row['user_min'], axis=1)
    return predictions


def merging_predicted_date_to_result_data(results,
                                          feuture_orders,
                                          customer_indicator,
                                          time_indicator,
                                          amount_indicator):
    results = results.rename(columns={"prediction_values": amount_indicator, 'pred_order_seq': 'order_seq_num'})
    results = pd.merge(results,
                       feuture_orders[[customer_indicator, 'order_seq_num', time_indicator]],
                       on=[customer_indicator, 'order_seq_num'], how='left')
    results['data_type'] = 'prediction'
    data_columns = [customer_indicator, 'order_seq_num', time_indicator, amount_indicator, 'data_type']
    return results[data_columns]


def merge_0_result_at_time_period(data,
                                  max_date,
                                  time_period,
                                  customer_indicator,
                                  time_indicator,
                                  amount_indicator):
    last_time_period_date = max_date + datetime.timedelta(days=convert_time_preiod_to_days(time_period))
    result = pd.DataFrame(data[customer_indicator].unique()).rename(columns={0: customer_indicator})
    result['order_seq_num'] = 1
    result[amount_indicator] = 0
    result['data_type'] = 'prediction'
    result[time_indicator] = last_time_period_date
    return result


reshape_3 = lambda x: x.reshape((x.shape[0], x.shape[1], 1))
reshape_2 = lambda x: x.reshape((x.shape[0], 1))


def split_data(Y, X, params):
    split_size = int(params['split_ratio'] * len(X))
    x_train = reshape_3(X[:split_size].values)
    y_train = reshape_2(Y[:split_size].values)
    x_test = reshape_3(X[split_size:].values)
    y_test = reshape_2(Y[split_size:].values)
    return {'x_train': x_train, 'y_train': y_train, 'x_test': x_test, 'y_test': y_test}


def drop_calculation(df, parameters, is_prediction=False):
    data_count = len(df)
    to_drop = max((parameters['tsteps'] - 1), (parameters['lahead'] - 1))
    df = df[to_drop:]
    if not is_prediction:
        if df.shape[0] > parameters['batch_size']:
            to_drop = df.shape[0] % parameters['batch_size']
        if to_drop > 0:
            df = df[:-1 * to_drop]
    return df


def arrange__data_for_model(df, f, parameters, is_prediction=False):
    try:
        y = df[f].rolling(window=parameters['tsteps'], center=False).mean()
    except Exception as e:
        print(df[f].head())
    x = pd.DataFrame(np.repeat(df[f].values, repeats=parameters['lag'], axis=1))
    shift_day = int(parameters['lahead'] / parameters['lag'])
    if parameters['lahead'] > 1:
        for i, c in enumerate(x.columns):
            x[c] = x[c].shift(i * shift_day)  # every each same days of shifted
    x = drop_calculation(x, parameters, is_prediction=is_prediction)
    y = drop_calculation(y, parameters, is_prediction=is_prediction)
    return split_data(y, x, parameters) if not is_prediction else reshape_3(x.values)


def convert_time_preiod_to_days(time_period):
    if time_period == 'day':
        return 1
    if time_period == 'year':
        return 365
    if time_period == 'month':
        return 30
    if time_period == 'week':
        return 7
    if time_period == '2*week':
        return 14
    if time_period == '2*month':
        return 60
    if time_period == 'quarter':
        return 90


def check_model_exists(path, model_name, time_period):
    current_date = datetime.datetime.strptime(get_current_day(replace=False), "%Y-%m-%d")
    _prev_model_day_diff = 0
    day_range_for_model_training = convert_time_preiod_to_days(time_period)
    prev_model = None
    for m in listdir(dirname(join(path, ""))):
        if "_".join(m.split("_")[:-2]) == model_name:
            _date_str = m.split("_")[-2]
            _date_str = "-".join([_date_str[0:4], _date_str[4:6], _date_str[6:]])
            _date = datetime.datetime.strptime(_date_str, "%Y-%m-%d")
            if abs((_date - current_date).total_seconds()) / 60 / 60 / 24 < day_range_for_model_training:
                if m.split("_")[-1].split(".")[0] == time_period:
                    prev_model = m
    return prev_model


def model_path(directory, model_name, time_period):
    return join(directory, model_name + "_" + get_current_day() + "_" + time_period.replace(" ", "") + ".json")


def weights_path(directory, model_name, time_period):
    return join(directory, model_name + "_" + get_current_day() + "_" + time_period.replace(" ", "") + ".h5")


def get_tuning_params(parameter_tuning, params):
    """
      activation: relu_tanh
      batch_size:
        - 5120
        - 10240
        - 20480
      epochs:  1000*40000
      units:
        - 8
        - 16
        - 32
        - 64

      drop_out_ratio': 0.1*0.5

    :param parameter_tuning:
    :param params:
    :param job:
    :return:
    """
    arrays = []
    hyper_params = {}
    for p in parameter_tuning:
        if type(parameter_tuning[p]) in [dict, list]:
            hyper_params[p] = parameter_tuning[p]
        if type(parameter_tuning[p]) == str:
            if "*" in parameter_tuning[p]:
                # e.g. 0.1*0.5 return [0.1, 0.2, ..., 0.5] or 0.1*0.5*0.05 return [0.1, 0.15, ..., 0.5]
                _splits = parameter_tuning[p].split("*")
                if len(_splits) == 2:
                    hyper_params[p] = np.arange(float(_splits[0]), float(_splits[1]), float(_splits[0])).tolist()
                if len(_splits) == 3:
                    hyper_params[p] = np.arange(float(_splits[0]), float(_splits[1]), float(_splits[2])).tolist()
                hyper_params[p] = [str(c) for c in hyper_params[p]]
            else:  # e.g. relu_tanh or relu
                hyper_params[p] = parameter_tuning[p].split("_")
    for p in hyper_params:
        if p not in ['activation', 'loss']:
            if p not in ['kernel_size', 'pool_size', 'max_pooling_unit', 'lstm_units',
                         'num_layers', 'units', 'batch_size']:
                hyper_params[p] = [float(c) for c in hyper_params[p]]
            else:
                if p != 'num_layers':
                    hyper_params[p] = [int(c) for c in hyper_params[p]]
                else:
                    hyper_params[p] = {c: int(hyper_params[p][c]) for c in hyper_params[p]}

    for p in params:
        if p not in list(hyper_params.keys()):
             hyper_params[p] = params[p]
    return hyper_params


def get_results(directory, time_period):
    results = pd.DataFrame()
    result_files = [f for f in listdir(dirname(join(directory, ""))) if f.split("_")[0] == "results"]
    parsed_date = [parse(i.split("_")[2]) for i in result_files]
    detected_file = None
    if len(parsed_date) != 0:
        current_date = min([parse(i.split("_")[2]) for i in result_files])
        date = current_date - datetime.timedelta(days=convert_time_preiod_to_days(time_period))
        detected_file = None
        for f in result_files:
            f_split = f.split("_")
            if f_split[1] == time_period:
                if parse(f_split[3].split(".")[0]) >= date:
                    date = parse(f_split[3].split(".")[0])
                    detected_file = f
    if detected_file is not None:
        results = pd.read_csv(join(directory, detected_file))
    return results


def check_for_previous_predicted_clv_results(results,
                                             path,
                                             time_period,
                                             time_indicator,
                                             customer_indicator,
                                             ):
    prev_result = get_results(path, time_period)
    if len(prev_result) != 0 and len(results) != 0:
        prev_result['same_order'] = True
        prev_result[time_indicator] = prev_result[time_indicator].apply(lambda x: str(x))
        results[time_indicator] = results[time_indicator].apply(lambda x: str(x))
        prev_result = pd.merge(prev_result,
                               results[[customer_indicator, time_indicator]],
                               on=[customer_indicator, time_indicator], how='left')
        prev_result = prev_result.query("same_order != same_order").drop('same_order', axis=1)
    results = pd.concat([prev_result, results])
    return results


def batch_size_hp_ranges(client_sample_sizes, num_of_customers, hyper_params):
    """
    For the computational cost and model of bias/variance problem related to other parameters of the NN.
    !!! optimum_batch_size: !!!
    client_sample_sizes are number of customers orders.
    The most frequent order count will be the business of the average of order count per user.
    batch size must be the  most frequent order count in order to
    make the predictive model more representative for customers.
    This will only become more affective when batch_size is assigned as The most frequent order count.
    That is why optimum_batch_size is The most frequent order count.

    additional to batch_size hyper parameters, optimum_batch_size and number of unique customers must be tunned.
    :param client_sample_sizes: number of row count per user
    :param num_of_customers: number ıf unique customer count
    :return: range of batch sizes, average_customer_batch
    """
    (unique, counts) = np.unique(client_sample_sizes, return_counts=True)
    optimum_batch_size = int(sorted(zip(counts, unique))[-1][1])
    average_customer_batch = int(num_of_customers - (num_of_customers % optimum_batch_size))
    hyper_params['batch_sizes'] = sorted(hyper_params['batch_sizes'] +
                                         [optimum_batch_size, average_customer_batch], reverse=True)
    return hyper_params, average_customer_batch


class OptimumLagDecision:
    def __init__(self, data, customer_indicator, time_indicator, params, directory):
        self.data = data
        self.customer_indicator = customer_indicator
        self.time_indicator = time_indicator
        self.directory = directory
        self.minimum_batch_size_count = params['batch_size']
        self.split_ratio = params['split_ratio']
        self.lag_range = list(range(1, 4))
        self.train, self.test, self.prediction = [], [], []
        self.model = None
        self.best_lag = 1
        self.min_rmse = 1000000000000

    def collect_data(self):
        self.data['order_seq_num'] = self.data.sort_values(by=[self.customer_indicator,
                                                               self.time_indicator]
                                                           ).groupby([self.customer_indicator]).cumcount() + 1
        self.data = self.data.sort_values(by='order_seq_num', ascending=True).groupby('order_seq_num').agg(
            {"time_diff": "mean"}).reset_index()

    def split_data(self):
        self.train = list(
            self.data[self.data['order_seq_num'] < int(max(self.data['order_seq_num']) * self.split_ratio)][
                'time_diff'])
        self.test = list(
            self.data[self.data['order_seq_num'] >= int(max(self.data['order_seq_num']) * self.split_ratio)][
                'time_diff'])

    def build_arima_model(self, p):
        self.prediction = []
        for i in self.test:
            try:
                self.model = ARIMA(self.train, order=(p, 1, 1))
                self.model = self.model.fit()
                self.prediction.append(self.cal_residuals(i, self.model.forecast(steps=1)[0]))
            except Exception as e:
                print(e)
            self.train += [i]
        return sqrt(sum(self.prediction))

    def cal_residuals(self, actual, prediction):
        return pow(abs(actual - prediction), 2)

    def find_optimum_lag(self):
        if check_for_existing_parameters(self.directory, 'next_purchase') is None:
            self.collect_data()
            self.split_data()
            for lag in self.lag_range:
                _rmse = self.build_arima_model(lag)
                print("lag :", lag, "RMSE :", _rmse)
                if self.min_rmse > _rmse:
                    self.min_rmse = _rmse
                    self.best_lag = lag
            print("optimum lag :", self.best_lag)






import warnings
from pandas import DataFrame, concat
import os
import shutil
from itertools import product
import glob
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '4'

from tensorflow.keras.layers import Dense, LSTM, Input, BatchNormalization
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.initializers import Ones
from tensorflow.keras.models import Model
from tensorflow.keras.models import model_from_json
from kerastuner.tuners import RandomSearch
from kerastuner.engine.hyperparameters import HyperParameters

try:
    from functions import *
    from configs import hyper_conf, accept_threshold_for_loss_diff, parameter_tuning_trials
    from data_access import *
except Exception as e:
    from .functions import *
    from .configs import hyper_conf, accept_threshold_for_loss_diff, parameter_tuning_trials
    from .data_access import *


def model_from_to_json(path=None, weights_path=None, model=None, is_writing=False):
    if is_writing:
        model_json = model.to_json()
        with open(path, "w") as json_file:
            json_file.write(model_json)
        model.save_weights(weights_path)
    else:
        json_file = open(path, 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        model = model_from_json(loaded_model_json)
        model.load_weights(weights_path)
        return model


def updating_hyper_parameters_related_to_data():
    return None


def get_params(params, comb):
    count = 0
    for p in params:
        _p = type(params[p])(comb[count])
        params[p] = _p
        count += 1
    return params


class TrainLSTM:
    def __init__(self,
                 date=None,
                 time_indicator=None,
                 order_count=None,
                 data_source=None,
                 data_query_path=None,
                 time_period=None,
                 directory=None,
                 customer_indicator=None,
                 amount_indicator=None):
        self.directory = directory
        self.customer_indicator = customer_indicator
        self.time_indicator = time_indicator
        self.amount_indicator = amount_indicator
        self.params = hyper_conf('next_purchase')
        self.hyper_params = get_tuning_params(hyper_conf('next_purchase_hyper'), self.params)
        self.optimized_parameters = {}
        self._p = None
        self.order_count = order_count
        self.time_period = time_period
        self.data, self.features, self.c_min_max, self.params = data_manipulation_np(date=date,
                                                                                     feature=amount_indicator,
                                                                                     time_indicator=time_indicator,
                                                                                     data_source=data_source,
                                                                                     data_query_path=data_query_path,
                                                                                     customer_indicator=customer_indicator,
                                                                                     params=self.params,
                                                                                     time_period=time_period,
                                                                                     directory=directory)

        self.hp = HyperParameters()
        self.model_data = {}
        self.input, self.model = None, None
        self.model = check_model_exists(self.directory, "trained_next_purchase_model", self.time_period)
        self.result = None, None
        self.residuals, self.anomaly = [], []
        self.customers = list(self.data[customer_indicator].unique())
        self.results = DataFrame()
        self.get_actual_value = lambda _min, _max, _value: ((_max - _min) * _value) + _min
        self.max_date = max(self.data[self.time_indicator])
        self.future_date = self.max_date + datetime.timedelta(days=convert_time_preiod_to_days(self.time_period))
        self.model_data = {"x_train": None, "y_train": None, "x_test": None, "y_test": None}
        self.client_sample_sizes = []
        self.optimum_batch_size = 32

    def get_model_data(self, customer):
        _data = self.data[self.data[self.customer_indicator] == customer]
        try:
            data = arrange__data_for_model(df=_data, f=[self.features], parameters=self.params)
            if data['x_train'].shape[0] != 0:
                model_data[customer] = {}
                model_data[customer] = data
        except Exception as e:
            print(e)

    def data_preparation(self):
        global model_data
        model_data = {}
        print("number of customers :", len(self.customers))
        execute_parallel_run(self.customers, self.get_model_data, arguments=None, parallel=16)
        try_count = 0
        while try_count < 20:
            try:
                self.model_data = {"x_train": None, "y_train": None, "x_test": None, "y_test": None}
                self.client_sample_sizes = []
                for c in model_data:
                    if self.model_data['x_train'] is not None:
                        for _split in ['x_train', 'y_train', 'x_test', 'y_test']:
                            self.model_data[_split] = np.concatenate([self.model_data[_split], model_data[c][_split]])
                        self.client_sample_sizes.append(model_data[c]['x_train'].shape[0])
                    else:
                        self.client_sample_sizes.append(model_data[c]['x_train'].shape[0])
                        self.model_data = model_data[c]
                try_count = 20
            except Exception as e:
                print(e)
                try_count += 1

    def build_parameter_tuning_model(self, hp):
        self.input = Input(shape=(self.model_data['x_train'].shape[1], 1))
        lstm = LSTM(int(hp.Choice('units', self.hyper_params['units'])),
                    bias_initializer=Ones(),
                    kernel_initializer=Ones(),
                    use_bias=False,
                    activation=hp.Choice('activation', self.hyper_params['activation']),
                    dropout=0.1
                    )(self.input)
        lstm = BatchNormalization()(lstm)
        lstm = Dense(1)(lstm)
        model = Model(inputs=self.input, outputs=lstm)
        model.compile(loss='mae',
                      optimizer=Adam(lr=hp.Choice('lr', self.hyper_params['lr'])),
                      metrics=['mae'])
        return model

    def build_model(self):
        self.input = Input(shape=(self.model_data['x_train'].shape[1], 1))
        # LSTM layer
        lstm = LSTM(self.params['units'],
                    batch_size=self.params['batch_size'],
                    bias_initializer=Ones(),
                    kernel_initializer=Ones(),
                    use_bias=False,
                    recurrent_activation=self.params['activation'],
                    dropout=0.1
                    )(self.input)
        lstm = BatchNormalization()(lstm)
        lstm = Dense(1)(lstm)
        self.model = Model(inputs=self.input, outputs=lstm)
        self.model.compile(loss='mae', optimizer=Adam(lr=self.params['lr']), metrics=['mae'])

    def learning_process(self, save_model=True, history=False):
        if history:
            history = self.model.fit(self.model_data['x_train'],
                                     self.model_data['y_train'],
                                     batch_size=self.params['batch_size'],
                                     epochs=int(self.params['epochs']),
                                     verbose=1,
                                     validation_data=(self.model_data['x_test'], self.model_data['y_test']),
                                     shuffle=True)
        else:
            print("*"*5, "Fit Next Purchase Model", "*"*5)
            self.model.fit(self.model_data['x_train'],
                           self.model_data['y_train'],
                           batch_size=self.params['batch_size'],
                           epochs=int(self.params['epochs']),
                           verbose=1,
                           validation_data=(self.model_data['x_test'], self.model_data['y_test']),
                           shuffle=True)
        if save_model:
            model_from_to_json(path=model_path(self.directory,
                                               "trained_next_purchase_model", self.time_period),
                               weights_path = weights_path(self.directory,
                                               "trained_next_purchase_model", self.time_period),
                               model=self.model,
                               is_writing=True)

        if history:
            return history

    def train_execute(self):
        print("*"*5, "Next purchase train model process ", "*"*5)
        if self.model is None:
            self.data_preparation()
            self.parameter_tuning()
            self.build_model()
            self.learning_process()
        else:
            self.model = model_from_to_json(path=model_path(self.directory,
                                                            "trained_next_purchase_model", self.time_period),
                                            weights_path=weights_path(self.directory,
                                                                      "trained_next_purchase_model", self.time_period))
            print("Previous model already exits in the given directory  '" + self.directory + "'.")

    def prediction_date_add(self, data, pred_data, pred):
        max_date = max(data[self.time_indicator]) if len(pred_data) == 0 else max(pred_data[self.time_indicator])
        if self.time_period != 'hour':
            _predicted_date = max_date + datetime.timedelta(minutes=int(round(pred)))
        if self.time_period != 'day':
            _predicted_date = max_date + datetime.timedelta(hours=int(round(pred)))
        if self.time_period not in ['day', 'hour']:
            _predicted_date = max_date + datetime.timedelta(days=int(round(pred)))
        return _predicted_date

    def predicted_date_in_range_decision(self, end, _pred_data, _predicted_date, customer, _pred_actual, _pred):
        if _predicted_date < end:
            _pred_data = concat([_pred_data, DataFrame([{self.time_indicator: _predicted_date,
                                                         self.customer_indicator: customer,
                                                         'time_diff': _pred_actual,
                                                         'time_diff_norm': _pred}])])
        return _pred_data

    def calculate_prediction(self, data, _pred_data, user_min, user_max):
        try:
            x = data_for_customer_prediction(data, _pred_data, self.params)
            _pred = self.model.predict(x)[0][-1]
            _pred_actual = self.get_actual_value(_min=user_min, _max=user_max, _value=_pred)
        except Exception as e:
            print(e)
            _pred_actual, _pred = 0, 0
        return _pred_actual, _pred

    def prediction_per_customer(self,  customer):
        warnings.simplefilter("ignore")
        user_min = self.temp_data[customer]['user_min']
        user_max = self.temp_data[customer]['user_max']
        data = pd.DataFrame(self.temp_data[customer]['customer_transactions']
                            ).rename(columns={0: self.time_indicator,
                                              1: self.amount_indicator,
                                              2: 'time_diff_norm',
                                              3: 'time_diff'})
        _pred_data, prediction_data[customer] = DataFrame(), DataFrame()
        _predicted_date = self.future_date - datetime.timedelta(days=1)
        _pred_actual, _pred = self.calculate_prediction(data, _pred_data, user_min, user_max)
        if self.max_date < self.prediction_date_add(data, pd.DataFrame(), (_pred_actual * 3) + 0.05):
            while self.max_date < _predicted_date < self.future_date:
                try:
                    _pred_actual, _pred = self.calculate_prediction(data, _pred_data, user_min, user_max)
                    if int(round(_pred_actual)) != 0:
                        _predicted_date = self.prediction_date_add(data, _pred_data, _pred_actual)
                        _pred_data = self.predicted_date_in_range_decision(self.future_date,
                                                                           _pred_data,
                                                                           _predicted_date,
                                                                           customer,
                                                                           _pred_actual,
                                                                           _pred)
                    else:
                        _predicted_date = self.future_date + datetime.timedelta(days=1)
                except Exception as e:
                    _predicted_date = self.future_date + datetime.timedelta(days=1)

            if len(_pred_data) != 0:
                try:
                    _pred_data = _pred_data[(_pred_data[self.time_indicator] >= self.max_date) &
                                            (_pred_data[self.time_indicator] < self.future_date) &
                                            (_pred_data[self.time_indicator] == _pred_data[self.time_indicator])]
                except Exception as e:
                    print(e)
            prediction_data[customer] = _pred_data

    def create_prediction_data(self):
        self.temp_data = self.data.groupby(self.customer_indicator).agg(
            {self.time_indicator: lambda x: list(x),
             self.amount_indicator: lambda x: list(x),
             "time_diff_norm": lambda x: list(x), 'time_diff': lambda x: list(x)}).reset_index()
        self.temp_data['customer_transactions'] = self.temp_data.apply(
            lambda row: list(zip(row[self.time_indicator], row[self.amount_indicator],
                                 row['time_diff_norm'], row['time_diff'])), axis=1)
        self.temp_data = self.temp_data.drop([self.time_indicator, self.amount_indicator], axis=1)
        self.temp_data = pd.merge(self.temp_data, self.c_min_max, on=self.customer_indicator, how='left')
        self.temp_data = {i[1]: i[0] for i in zip(self.temp_data.set_index(self.customer_indicator).to_dict('results'),
                                                  list(self.temp_data[self.customer_indicator]))}

    def parallel_prediction(self):
        try:
            os.mkdir(join(self.directory, "temp_next_purchase_results"))
        except Exception as e:
            print(e)
            print("recreating 'temp_next_purchase_results' folder ...")
            shutil.rmtree(join(self.directory, "temp_next_purchase_results", ""))
            os.mkdir(join(self.directory, "temp_next_purchase_results"))

        iters = int(len(self.customers) / 256) + 1
        for i in range(iters):
            print("main iteration :", str(i), " / ", str(iters))
            _sample_customers = get_iter_sample(self.customers, i, iters, 256)
            global prediction_data
            prediction_data = {}
            execute_parallel_run(_sample_customers, self.prediction_per_customer, arguments=None, parallel=8)
            _result = pd.DataFrame()
            for c in _sample_customers:
                try:
                    _result = concat([_result, prediction_data[c]])
                except Exception as e:
                    print(c)
            _result.to_csv(join(self.directory, "temp_next_purchase_results", str(_sample_customers[0] + "_data.csv")),
                           index=False)

    def create_prediction_result_data(self):
        print("merge predicted data ...")
        _import_files = glob.glob(join(self.directory, "temp_next_purchase_results", "*.csv"))
        li = []
        for f in _import_files:
            try:
                _df = pd.read_csv(f, index_col=None, header=0)
                li.append(_df)
            except Exception as e:
                print(e)
        self.results = concat(li, axis=0, ignore_index=True)
        self.results[self.time_indicator] = self.results[self.time_indicator].apply(lambda x: convert_str_to_day(x))
        print(self.results.head())
        shutil.rmtree(join(self.directory, "temp_next_purchase_results", ""))
        self.results = self.results[(self.results[self.time_indicator] > self.max_date) &
                                    (self.results[self.time_indicator] < self.future_date)]
        print(self.results.head())
        print("number of predicted customers :", len(self.results[self.customer_indicator].unique()))

    def prediction_execute(self):
        print("*"*5, "PREDICTION", 5*"*")
        print("number of users :", len(self.customers))
        if self.model is not None:
            self.model = model_from_to_json(path=model_path(self.directory,
                                                            "trained_next_purchase_model", self.time_period),
                                            weights_path=weights_path(self.directory,
                                                                      "trained_next_purchase_model", self.time_period))
        self.results = self.data[[self.time_indicator, 'time_diff', 'time_diff_norm', self.customer_indicator]]
        self.create_prediction_data()
        self.parallel_prediction()
        self.create_prediction_result_data()

    def initialize_keras_tuner(self):
        """
        Parameter tuning process is triggered via Keras-Turner Library.
        However, batch_size and epoch parameters of optimization are created individually.
        """

        self.hyper_params, self.optimum_batch_size = batch_size_hp_ranges(client_sample_sizes=self.client_sample_sizes,
                                                                          num_of_customers=len(self.customers),
                                                                          hyper_params=self.hyper_params)
        tuner = RandomSearch(
            self.build_parameter_tuning_model,
            max_trials=parameter_tuning_trials,
            hyperparameters=self.hp,
            allow_new_entries=True,
            objective='loss')
        tuner.search(x=self.model_data['x_train'],
                     y=self.model_data['y_train'],
                     epochs=5,
                     batch_size=self.optimum_batch_size,
                     verbose=1,
                     validation_data=(self.model_data['x_test'], self.model_data['y_test']))
        for p in tuner.get_best_hyperparameters()[0].values:
            if p in list(self.params.keys()):
                self.params[p] = tuner.get_best_hyperparameters()[0].values[p]

    def batch_size_and_epoch_tuning(self):
        """
        It finds the optimum batch_size and epoch.
        Each chosen epoch and batch_size of model created via model.
        Last epoch of loss is lower than 'accept_threshold_for_loss_diff' decision be finalized.
        Epoch will test ascending format (e.g. 4, 8, 16, ..., 1024) in order to minimize time consumption
        of both parameter tuning and model creation.
        Batch size will test descending format (e.g. 1024, 512, ...4) in order to minimize time consumption
        of both parameter tuning and model creation.
        """

        counter = 0
        optimum_epoch_process_done = False
        epoch_bs_combs = list(product(sorted(self.hyper_params['batch_sizes'], reverse=True),
                                      sorted(self.hyper_params['epochs'])
                                      ))
        loss_values = []
        while not optimum_epoch_process_done:
            self.params['epochs'] = int(epoch_bs_combs[counter][1])
            self.params['batch_size'] = int(epoch_bs_combs[counter][0])
            self.build_model()
            _history = self.learning_process(save_model=False, history=True)
            if _history.history['loss'][-1] < accept_threshold_for_loss_diff:
                optimum_epoch_process_done = True
            loss_values.append({"epochs": self.params['epochs'],
                                "batch_size": self.params['batch_size'],
                                "loss": _history.history['loss'][-1]})
            counter += 1
            if counter >= parameter_tuning_trials:
                optimum_epoch_process_done = True
                loss_values_df = pd.DataFrame(loss_values).sort_values(by='loss', ascending=True)
                self.params['epochs'] = list(loss_values_df['epochs'])[0]
                self.params['batch_size'] = list(loss_values_df['batch_size'])[0]

    def remove_keras_tuner_folder(self):
        """
        removing keras tuner file. while you need to update the parameters it will affect rerun the parameter tuning.
        It won`t start unless the folder has been removed.
        """

        try:
            shutil.rmtree(
                join(abspath(__file__).split("next_purchase_model.py")[0].split("clv")[0][:-1], "clv_prediction",
                     "untitled_project"))
        except Exception as e:
            print(" Parameter Tuning Keras Turner dummy files have already removed!!")

    def update_tuned_parameter_file(self):
        """
        tuned parameters are stored at 'test_parameters.yaml.' in given 'export_path' argument.
        """

        if check_for_existing_parameters(self.directory, 'purchase_amount') is not None:
            _params = read_yaml(self.directory, "test_parameters.yaml")
            _params['next_purchase'] = self.params
        else:
            print("Non of parameter tuning for both Model has been observed.")
            _params = {'next_purchase': self.params}
        write_yaml(self.directory, "test_parameters.yaml", _params, ignoring_aliases=True)

    def parameter_tuning(self):
        if check_for_existing_parameters(self.directory, 'next_purchase') is None:
            self.initialize_keras_tuner()
            self.batch_size_and_epoch_tuning()
            self.remove_keras_tuner_folder()
            self.update_tuned_parameter_file()

        else:
            self.params = check_for_existing_parameters(self.directory, 'next_purchase')







import schedule
import datetime
import time
import argparse


try:
    from main import main
    from utils import convert_date, read_yaml, write_yaml
except Exception as e:
    from .main import main
    from .utils import convert_date, read_yaml, write_yaml


def convert_to_day_hour(x):
    return datetime.datetime.strptime(str(x)[0:13], "%Y-%m-%d %H")


def convert_to_day(x):
    return datetime.datetime.strptime(str(x)[0:10], "%Y-%m-%d")


def get_schedule(time_period):
    if time_period not in ['minute', 'hour', 'week']:
        return {'mondays': schedule.every().monday,
                'tuesdays': schedule.every().tuesday,
                'wednesdays': schedule.every().wednesday,
                'thursdays': schedule.every().thursday,
                'fridays': schedule.every().friday,
                'saturdays': schedule.every().saturday,
                'sundays': schedule.every().sunday,
                'day': schedule.every().day
                }[time_period]
    if time_period == 'week':
        return schedule.every().week
    if time_period == 'hour':
        print("initial time :", str(datetime.datetime.now())[11:16])
        return schedule.every(1).hours.at(str(datetime.datetime.now())[11:16])


def update_date(args):
    iteration, time_schedule = args['iteration'], args['time_schedule']
    if time_schedule == 'hour':
        if len(str(args['arguments']['date'])) == 10:
            args['arguments']['date'] = args['arguments']['date'] + ' 00'
        date = convert_to_day_hour(args['arguments']['date'])
    else:
        date = convert_to_day(args['arguments']['date'])
    if iteration != 0:
        if time_schedule == 'hour':
            date = date + datetime.timedelta(hours=iteration)
        if time_schedule == 'Daily':
            date = date + datetime.timedelta(days=iteration)
        if time_schedule == 'week':
            date = date + datetime.timedelta(days=int(7 * iteration))
    args['arguments']['date'] = str(date)[0:13]
    write_yaml(args['arguments']['export_path'], 'schedule_' + args['arguments']['job'] + '.yaml',
               {'arguments': args['arguments'],
                'time_schedule': args['time_schedule'],
                'iteration': args['iteration'] + 1
                })
    print("CLV Prediction Date :", args['arguments']['date'])


def run_ab_test(directory, job):
    arguments = read_yaml(directory, 'schedule_' + job + '.yaml')
    update_date(arguments)
    main(**arguments['arguments'])


def check_for_first_running(args):
    """
    before initialize the periodically run, it must be trained.
    If there are the trained model stored in the given directory, process will be skipped.
    args: arguments,  in order to initialize the train or prediction process.
    :return: None
    """
    if args['job'] == 'train':
        main(**args)


def decision_of_time_sleep(time_schedule):
    if time_schedule == 'hour':
        return 60
    if time_schedule == 'day':
        return 600
    if time_schedule == 'week':
        return 3600
    if time_schedule == 'month':
        return 18000
    if time_schedule not in ['hour', 'day', 'week', 'month']:
        return 1000


def create_job(arguments, time_schedule):
    time_schedule = time_schedule
    args = arguments
    check_for_first_running(args)
    _sch = get_schedule(time_schedule)
    _sch.do(run_ab_test, arguments['export_path'], arguments['job'])
    print(_sch)
    time_sec_wait = decision_of_time_sleep(time_schedule)
    while True:
        schedule.run_pending()
        time.sleep(time_sec_wait)
        print("waiting ...")


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-TS", "--time_schedule", type=str,
                        help="""
                        
                        """)
    parser.add_argument("-J", "--job", type=str,
                        help="""

                        """,
                        )
    parser.add_argument("-OC", "--order_count", type=str,
                        help="""

                        """,
                        )
    parser.add_argument("-CI", "--customer_indicator", type=str,
                        help="""

                        """,
                        )
    parser.add_argument("-AI", "--amount_indicator", type=str,
                        help="""

                        """,
                        )
    parser.add_argument("-DS", "--data_source", type=str,
                        help="""
                        AWS RedShift, BigQuery, PostgreSQL, csv, json files can be connected to system
                        """,
                        required=True)
    parser.add_argument("-DQP", "--data_query_path", type=str,
                        help="""
                        if there is file for data importing;
                            must be the path (e.g /../.../ab_test_raw_data.csv)
                        if there is ac- connection such as PostgreSQL / BigQuery
                            query must at the format "SELECT++++*+++FROM++ab_test_table_+++"
                        """,
                        required=True)
    parser.add_argument("-TI", "--time_indicator", type=str,
                        help="""
                        This can only be applied with date. It can be hour, day, week, week_part, quarter, year, month.
                        Individually time indicator checks the date part is significantly 
                        a individual group for data set or not.
                        If it is uses time_indicator as a  group
                        """,
                        )
    parser.add_argument("-EP", "--export_path", type=str,
                        help="""
                        This shows us to the time period if AB Test is running sequentially
                        """,
                        )
    arguments = parser.parse_args()
    clv_arguments = {'test_groups': arguments.test_groups,
                         'groups': arguments.groups, 'date': arguments.date,
                         'feature': arguments.feature, 'data_source': arguments.data_source,
                         'data_query_path': arguments.data_query_path, 'time_period': arguments.time_period,
                         "export_path": arguments.export_path}

    create_job(clv_arguments, arguments.time_schedule)